#pragma warning(disable:4704)

#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <dos.h>

#include "aspi.h"

//  Local definitions.

#define DOS_INT 0x21			// DOS interrupt

//  Local prototypes.

static SHORT GenDOS( WORD axreg, WORD bxreg, WORD cxreg, LPVOID iopb,
		   LPWORD errstat);

//  Locals.

static SHORT
  hASPI = -1,				// ASPI device handle
  ASPITimeOut = ASPI_DEFAULT_TIMEOUT;	// default to timeout

static time_t
  StartTime;				// starting time for operation


//  Structure to call an ASPI service.

static union
{
  VOID (FAR *ASPI_Entry)( LPVOID packet);
  struct
  {
    WORD ASPI_Offset;		// ASPI offset
    WORD ASPI_Segment;		// ASPI segment
  };
} ServiceASPI;


//  AccessASPI - Obtain/Drop ASPI access.
//  -------------------------------------
//
//  int AccessASPI( void)
//
//  Returns 0 or 1, depending on ASPI presence.
//

SHORT FAR AccessASPI( VOID)
{

  SHORT
    htemp;

  hASPI = -1;				// make sure of it
  if ( !GenDOS( 0x3d00, 0, 0, ASPI_DEV_NAME, addr(htemp)))
  { // got an open!
    GenDOS( 0x4402, htemp, 4, addr( ServiceASPI.ASPI_Offset), 0);
    GenDOS( 0x3e00, htemp, 0, 0, 0);
    hASPI = 0;				// say okay
    return 1;
  }
  return 0;
} // AccessASPI


//  CallASPI - Call ASPI with request.
//  ----------------------------------
//
//    We format the SRB here, but you have to pass one in.
//
//    Error status from SRB is returned.  -1 is returned if no ASPI access
//    has been obtained.

SHORT FAR CallASPI( LPVOID srb, BYTE func, BYTE adapt, BYTE flags)
{

  LPASPI_SRB
    header;	      // basic SRB header

  header = srb;				   // set location
  if ( hASPI < 0)
    return -1;				// no soap
  header->Command = func;
  header->Status = 0;
  header->Adapter = adapt;
  header->Flags = flags & ~ASPI_EXEC_NO_WAIT;
  header->Reserved = 0;
  (*(ServiceASPI.ASPI_Entry)) ( header);
  StartTime = time( NULL);
  if ( flags & ASPI_EXEC_NO_WAIT)
    return 1;				// always return busy
  else
  {
    while( PollASPI( srb) == 1);	// poll for completion
    return (header->Status == ASPI_STATUS_NONE) ? 0 : -1;
  }
} // CallASPI

//  ExecASPI - Execute an ASPI command.
//  -----------------------------------
//
//    More involved setup--waits for completion.
//
//    Note that we define an extra flag:  ASPI_EXEC_NO_WAIT
//
//    The nice part is that we invert the order of the bytes in all of
//    the appropriate CDB fields, so you don't have to.
//

SHORT FAR ExecASPI( LPVOID srb, BYTE sfunc, BYTE adapt, BYTE flags,
		   BYTE target, LPVOID Buffer, DWORD Buflen)
{

  LPASPI_EXECUTE
    execute;			    // execute command

//  We can fill in some additional information.

  if ( hASPI < 0)
    return -1;				// no soap
  execute = srb;
  execute->TargetID = target;
  execute->LUN = 0;		     // set target ID
  execute->SRBLink = NULL;	     // no link, ever
//  execute->PostRoutine = NULL;       // no post routine
  execute->CommandBlock.CDB6.Opcode = sfunc;  // command byte
  execute->SenseAlloc = ASPI_MAX_SENSE;	 // sense length
  if ( (sfunc >> 5) != 0)
    execute->CDBLength = 10;
  else
    execute->CDBLength = 6;
  execute->DataBuffer = Buffer;
  execute->DataAlloc = Buflen;	     // buffer size
  return CallASPI( srb, ASPI_CMD_EXECUTE, adapt, flags);
} // ExecASPI

//  ExecASPI - Execute an ASPI command.
//  -----------------------------------
//
//    More involved setup--waits for completion.
//
//    Note that we define an extra flag:  ASPI_EXEC_NO_WAIT
//
//    The nice part is that we invert the order of the bytes in all of
//    the appropriate CDB fields, so you don't have to.
//

SHORT FAR ResetASPI( LPVOID srb, BYTE adapt, BYTE target)
{

  LPASPI_RESET
    reset;			  // rest command

//  We can fill in some additional information.

  if ( hASPI < 0)
    return -1;				// no soap
  reset = srb;
  reset->TargetID = target;
  reset->LUN = 0;		   // set target ID
  reset->PostRoutine = NULL;	   // no post routine
  return CallASPI( srb, ASPI_CMD_RESET, adapt, 0);
} // ResetASPI

//  PollASPI - Poll ASPI for completion.
//  ------------------------------------
//
//    Returns 1 if busy, 0 if complete, -1 if timeout.
//
//    If a timeout occurs, a device reset is automatically issued.
//

SHORT FAR PollASPI( LPVOID srb)
{

  ASPI_ABORT
    Abort;				// abort command block
  LPASPI_SRB
    header;

  if ( hASPI < 0)
    return -1;				// no soap
  header = srb;

  if ( header->Status != ASPI_STATUS_BUSY)
    return 0;
  if( (SHORT) (time(NULL) - StartTime) < ASPITimeOut)
    return 1;

//  Whoops -- bad news, guys!

  memset( &Abort, 0, sizeof(Abort));
  Abort.SRB.Command = ASPI_CMD_ABORT;
  Abort.SRBPointer = srb;
  Abort.SRB.Adapter = header->Adapter;
  (*(ServiceASPI.ASPI_Entry)) ( &Abort);

  header->Status = ASPI_STATUS_ABORT;
  return -1;				// complete aborted command
} // PollASPI

//  SetASPITime - Set Timeout on ASPI commands.
//  -------------------------------------------
//
//    This routine just sets a global variable for time-outs.
//    A value of 0 seconds results in the default timeout.
//

SHORT FAR SetASPITime( SHORT secs)
{

  ASPITimeOut = secs ? secs : ASPI_DEFAULT_TIMEOUT;
  return ASPITimeOut;

} // SetASPITime


//    GenDOS - Issue a generic DOS function.
//    --------------------------------------
//
//    Takes (ax) (bx) and (cx) register contents and a pointer to
//    the parameter block (passed in dx).
//
//    Returns error status in *errstat.	 If error, return is -1,
//    otherwise 0.

static SHORT GenDOS( WORD axreg, WORD bxreg, WORD cxreg, LPVOID iopb,
		    LPWORD errstat)
{

  union _REGS
    regs;
  struct _SREGS
    segRegs;

  _segread( &segRegs);
  regs.x.ax = axreg;
  regs.x.bx = bxreg;
  regs.x.cx = cxreg;
  regs.x.dx = GET_FP_OFF( iopb);
  segRegs.ds = GET_FP_SEG( iopb);
  _intdosx( &regs, &regs, &segRegs);
  if ( errstat)
    *errstat = regs.x.ax;		// return error status if requested
  return regs.x.cflag ? -1 : 0;		// return completion status
}  // GenDOS

//  ASPIGet2 - Get two bytes from a field.
//  --------------------------------------
//
//    Input is pointer to field, return is WORD with field value set.
//

WORD FAR ASPIGet2( LPBYTE what)
{

  union
  {
    WORD wvalue;			// word value
    BYTE bvalue[2];			// byte value
  } decomp;

  decomp.bvalue[0] = what[1];
  decomp.bvalue[1] = what[0];
  return decomp.wvalue;			// return the value
} // ASPIGet2

//  ASPIGet3 - Get three bytes from a field.
//  ----------------------------------------
//
//    Input is pointer to field, return is DWORD with field value set.
//

DWORD FAR ASPIGet3( LPBYTE what)
{

  union
  {
    DWORD dvalue;			// dword value
    BYTE bvalue[4];			// byte value
  } decomp;

  decomp.bvalue[0] = what[2];
  decomp.bvalue[1] = what[1];
  decomp.bvalue[2] = what[0];
  decomp.bvalue[3] = 0;
  return decomp.dvalue;			// return the value
} // ASPIGet3

//  ASPIGet4 - Get four bytes from a field.
//  ----------------------------------------
//
//    Input is pointer to field, return is DWORD with field value set.
//

DWORD FAR ASPIGet4( LPBYTE what)
{

  union
  {
    DWORD dvalue;			// dword value
    BYTE bvalue[4];			// byte value
  } decomp;

  decomp.bvalue[0] = what[3];
  decomp.bvalue[1] = what[2];
  decomp.bvalue[2] = what[1];
  decomp.bvalue[3] = what[0];
  return decomp.dvalue;			// return the value
} // ASPIGet4

//  ASPIPut4 - Put four bytes to a field.
//  -------------------------------------
//
//    Input is dword value and pointer to field.
//

VOID FAR ASPIPut4( LPBYTE what, DWORD lvalue)
{

  union
  {
    DWORD dvalue;			// dword value
    BYTE bvalue[4];			// byte value
  } decomp;

  decomp.dvalue = lvalue;
  what[3] = decomp.bvalue[0];
  what[2] = decomp.bvalue[1];
  what[1] = decomp.bvalue[2];
  what[0] = decomp.bvalue[3];
  return;
} // ASPIPut4

//  ASPIPut3 - Put three bytes to a field.
//  --------------------------------------
//
//    Input is dword value and pointer to field.
//

VOID FAR ASPIPut3( LPBYTE what, DWORD lvalue)
{

  union
  {
    DWORD dvalue;			// dword value
    BYTE bvalue[4];			// byte value
  } decomp;

  decomp.dvalue = lvalue;
  what[2] = decomp.bvalue[0];
  what[1] = decomp.bvalue[1];
  what[0] = decomp.bvalue[2];
  return;
} // ASPIPut3

//  ASPIPut2 - Put two bytes to a field.
//  -------------------------------------
//
//    Input is word value and pointer to field.
//

VOID FAR ASPIPut2( LPBYTE what, WORD lvalue)
{

  union
  {
    WORD wvalue;			// dword value
    BYTE bvalue[4];			// byte value
  } decomp;

  decomp.wvalue = lvalue;
  what[1] = decomp.bvalue[0];
  what[0] = decomp.bvalue[1];
  return;
} // ASPIPut2

//  GetSenseInfo - Returns pointer to sense data.
//  ---------------------------------------------
//
//    Just a convienent way to get to the data.
//

LPSCSI_REQUEST_SENSE_RESULTS ASPIGetSenseInfo( LPASPI_EXECUTE SRB)
{
  return (LPSCSI_REQUEST_SENSE_RESULTS) ((LPBYTE)
	 &(SRB->CommandBlock.CDB6.Opcode) +
	 SRB->CDBLength);	// to the sense results
} // GetSenseInfo
