//  ASPI definitions.

#pragma pack(1)

#ifndef ASPI_INCLUDED

#include "commdefs.h"

#define ASPI_DEV_NAME "SCSIMGR$"	// device name
#define ASPI_MAX_CDB 10			// longest CDB
#define ASPI_MAX_SENSE 24		// longest sense data
#define ASPI_DEFAULT_TIMEOUT 30		// default command timeout

#include "scsi.h"

//  ASPI System Request Block (SRB).

typedef struct
{
  BYTE Command;				// command code
  BYTE Status;				// status of command
  BYTE Adapter;				// host adapter number
  BYTE Flags;				// request flags (see below)
  DWORD Reserved;			// reserved for future use
} ASPI_SRB, *PASPI_SRB, FAR *LPASPI_SRB;

//  ASPI Commands.

typedef enum
{
  ASPI_CMD_INQUIRE=0,			// Host Adapter Inquiry
  ASPI_CMD_GETTYPE,			// Get device type
  ASPI_CMD_EXECUTE,			// Execute SCSI I/O command
  ASPI_CMD_ABORT,			// Abort SCSI I/O command
  ASPI_CMD_RESET,			// Reset SCSI device
  ASPI_CMD_SETPARAM,			// Set adapter parameters
  ASPI_CMD_GETINFO,			// Get drive information
  ASPI_CMD_TOTAL			// total number of commands
} ASPI_COMMANDS;

//  ASPI Status values.

typedef enum
{
  ASPI_STATUS_BUSY	= 0,		// Request in progress
  ASPI_STATUS_NONE	= 1,		// No error
  ASPI_STATUS_ABORT	= 2,		// Command aborted by host
  ASPI_STATUS_ERROR	= 4,		// Command completed with error
  ASPI_STATUS_INVALID	= 128,		// Invalid SCSI request
  ASPI_STATUS_ADAPTER	= 129,		// Bad adapter number
  ASPI_STATUS_DEVICE	= 130		// Device not installed
} ASPI_STATUS;

//  ASPI command code 0; Host adapter inquiry.

typedef struct tagASPIInquire
{
  ASPI_SRB SRB;				// system request block
  BYTE Adapters;			// host adapter count
  BYTE HostID;				// ID of host adapter
  CHAR ManagerID[16];			// SCSI manager ID
  CHAR AdapterID[16];			// Host adapter ID
  BYTE AdapterParams[16];		// Adapter-unique parameters
} ASPI_INQUIRE, *PASPI_INQUIRE, FAR *LPASPI_INQUIRE;

//  ASPI command code 1; Get Device Type

typedef struct tagASPIGetType
{
  ASPI_SRB SRB;				// system request block header
  BYTE TargetID;			// SCSI address
  BYTE LUN;				// Logical Unit of device
  BYTE DeviceType;			// Peripheral device type
} ASPI_GETTYPE, *PASPI_GETTYPE, FAR *LPASPI_GETTYPE;

//  ASPI command code 2; Execute SCSI command

typedef struct tagASPIExecute
{
  ASPI_SRB SRB;				// system request block header
  BYTE TargetID;			// SCSI address
  BYTE LUN;				// Logical unit
  DWORD DataAlloc;			// Data Allocation length
  BYTE SenseAlloc;			// Sense Allocation length
  LPVOID DataBuffer;			// data buffer
  LPVOID SRBLink;			// SRB Link
  BYTE CDBLength;			// CDB Length
  BYTE AdapterStatus;			// Host adapter status
  BYTE TargetStatus;			// Target status
  VOID (FAR *PostRoutine)
       (struct tagASPIExecute FAR *);	// POST routine
  BYTE Reserved[34];			// Reserved for ASPI workspace
  CDB  CommandBlock;			// CDB Area
  BYTE Sense[ASPI_MAX_SENSE];		// Bytes for sense
} ASPI_EXECUTE, *PASPI_EXECUTE, FAR *LPASPI_EXECUTE;

//  Flags for execute command.

#define ASPI_EXEC_POST 1		// Use a POST routine
#define ASPI_EXEC_DIR_UNSPEC 0		// unspecified, no length check
#define ASPI_EXEC_DIR_READ   8		// read, check length
#define ASPI_EXEC_DIR_WRITE 16		// write, check length
#define ASPI_EXEC_DIR_NONE  24		// no data transfer
#define ASPI_EXEC_NO_WAIT 128		// do not wait for completion

//  ASPI command code 3; Abort SCSI command

typedef struct tagASPIAbort
{
  ASPI_SRB SRB;				// system request block header
  LPVOID SRBPointer;			// pointer to SRB to abort
} ASPI_ABORT, *PASPI_ABORT, FAR *LPASPI_ABORT;

//  ASPI command code 4; Reset SCSI device

typedef struct tagASPIReset
{
  ASPI_SRB SRB;				// system request block header
  BYTE TargetID;			// Target ID
  BYTE LUN;				// logical unit
  BYTE Reserved1[14];			// Reserved for expansion
  BYTE AdapterStatus;			// Host adapter status
  BYTE TargetStatus;			// Target status
  VOID (FAR *PostRoutine)
	(struct tagASPIReset FAR *);   // POST routine
  BYTE Reserved2[2];			// Reserved for ASPI workspace
} ASPI_RESET, *PASPI_RESET, FAR *LPASPI_RESET;

//  ASPI command code 5; Set host adapter parameters.

typedef struct tagASPISetParam
{
  ASPI_SRB SRB;				// system request block header
  BYTE AdapterParams[16];		// Adapter-unique parameters
} ASPI_SETPARAM, *PASPI_SETPARAM, FAR *LPASPI_SETPARAM;

//  ASPI command code 6; Get drive information.

typedef struct tagASPIGetInfo
{
  ASPI_SRB SRB;				// system request block header
  BYTE TargetID;			// Target ID
  BYTE LUN;				// logical unit
  BYTE DriveFlags;			// drive flags
  BYTE BIOSUnit;			// BIOS unit
  BYTE PreferedHeads;			// Prefered head translation
  BYTE PreferedSectors;			// Prefered sector translation
  BYTE Reserved[10];			// Reserved for expansion
} ASPI_GETINFO, *PASPI_GETINFO, FAR *LPASPI_GETINFO;

//  Prototypes.

SHORT FAR AccessASPI( VOID);
SHORT FAR CallASPI( LPVOID srb, BYTE func, BYTE adapt, BYTE flags);
SHORT FAR ExecASPI( LPVOID srb, BYTE sfunc, BYTE adapt, BYTE flags,
		  BYTE target, LPVOID Buffer, DWORD Buflen);
SHORT FAR ResetASPI( LPVOID srb, BYTE adapt, BYTE target);
SHORT FAR SetASPITime( SHORT secs);
SHORT FAR PollASPI( LPVOID srb);
WORD FAR ASPIGet2( LPBYTE what);
DWORD FAR ASPIGet3( LPBYTE what);
DWORD FAR ASPIGet4( LPBYTE what);
VOID FAR ASPIPut4( LPBYTE what, DWORD lvalue);
VOID FAR ASPIPut3( LPBYTE what, DWORD lvalue);
VOID FAR ASPIPut2( LPBYTE what, WORD lvalue);
LPSCSI_REQUEST_SENSE_RESULTS ASPIGetSenseInfo( LPASPI_EXECUTE SRB);
#define ASPIGetSenseKey(x) (ASPIGetSenseInfo(x)->SenseKey)
#define ASPIGetSenseByte(x) (ASPIGetSenseInfo(x)->SenseByte)
#define ASPI_INCLUDED
#endif
