//  COMMDEFS.H - Common type definitions.
//  -------------------------------------
//
//    Contains general definitions used by most routines.
//

#ifndef COMMON_DEFS
#define COMMON_DEFS 1
#define BYTE_DEFINED 1

#ifndef _H2INC
#ifndef addr
#define addr(x) (&(x))	      // basic address-of operator
#endif
#endif

#ifndef TRUE
#define TRUE 1		      // yes->true
#define FALSE 0		      // no->false
#endif

#ifndef NULL
#define NULL 0L		      // null value
#endif

#define FAR __far				   // define far pointer

//  A handle.

typedef int HANDLE, *PHANDLE, FAR *LPHANDLE;	  // handle

// 8-bit quantities.

#ifndef CHAR
typedef char CHAR;
#endif

typedef CHAR *PCHAR, FAR *LPCHAR;
typedef unsigned char UCHAR, *PUCHAR, FAR *LPUCHAR;
typedef unsigned char BYTE, *PBYTE, FAR *LPBYTE;

//  16-bit quantities.
#ifdef _H2INC
#define SHORT int
#else
typedef short SHORT;
#endif

typedef SHORT *PSHORT, FAR *LPSHORT;
typedef unsigned short int USHORT, *PUSHORT, FAR *LPUSHORT;

#ifdef _H2INC
#define WORD unsigned short int
#else
typedef unsigned short int WORD;
#endif
typedef WORD *PWORD, FAR *LPWORD;

//  32-bit quantities.

#ifndef LONG
typedef signed long LONG;
#endif
typedef signed long  *PLONG, FAR *LPLONG;
typedef unsigned long ULONG, *PULONG, FAR *LPULONG;
#ifdef _H2INC
#define DWORD unsigned long int
#else
typedef unsigned long int DWORD;
#endif
typedef DWORD *PDWORD, FAR *LPDWORD;

//  64-bit quantities.

typedef struct
{
  ULONG LowerPart;			// least significant 32 bits
  ULONG UpperPart;			// most significant 32 bits
} LARGE_INTEGER, *PLARGE_INTEGER, *LPLARGE_INTEGER;

//  Void types.

#ifndef VOID
typedef void VOID;
#endif
typedef VOID *PVOID, FAR *LPVOID;

//  Alias for const type.

#ifndef CONST
#define CONST const			// const keyword
#endif

//  Boolean (true/false) types.

#ifndef BOOLEAN
typedef unsigned short BOOLEAN;
#endif

typedef BOOLEAN *PBOOLEAN, FAR *LPBOOLEAN;

#ifndef _H2INC
#ifndef addr
#define addr(x) (&(x))			// address-of operator
#endif
#endif

#ifndef TRUE
#define TRUE 1				// affirmative
#define FALSE 0				// negative
#endif
#ifndef NULL
#define NULL 0
#endif

#ifndef _H2INC
//  The standard MS 'C' _FP_SEG and _FP_OFF have serious problems.  We
//  fix them here...

#ifdef _FP_SEG
#undef _FP_SEG
#undef _FP_OFF
#endif

#ifdef FP_SEG
#undef FP_SEG
#undef FP_OFF
#endif

__inline WORD GET_FP_SEG( LPVOID x)
{
   return (WORD) (((DWORD) x) >> 16);
} // GET_FP_SEG

#define GET_FP_OFF(x) ((WORD)((DWORD) ((LPVOID) (x))))

#ifdef MK_FP
#undef MK_FP
#endif

#define MAKE_FP(seg, offset) (LPVOID)(((ULONG)seg << 16)+(ULONG)(USHORT)offset)

#endif	  // _H2INC

//  Common interrupts.

#define MUX_INT 0x2f			// DOS multiplex interrupt
#define DOS_INT 0x21			// DOS main function interrupt
#define DISK_INT 0x13			// BIOS disk interrupt

#endif					// COMMON_DEFS
