//  SCSI Device definitions.

#pragma pack(1)

#ifndef SCSI_INCLUDED
#include "commdefs.h"

typedef struct
{
  BYTE Opcode;				// operation code
  union
  {
    struct
    {
      BYTE BEBlockLUN[3];		    // LUN/block ms nibble
      BYTE TransferCount;		    // Transfer length
    };
    struct
    {
      BYTE LUN;				// logical unit
      BYTE BETransferLength[3];		// transfer length
    };
  };
  BYTE Control;				// Control byte
} CDB_6;				// 6-byte CDB

typedef struct
{
  BYTE Opcode;				// operation code
  BYTE LUNRelAddr;			// LUN/Relative address flag
  BYTE BEBlock[4];			// block number (big endian)
  BYTE Reserved;			// reserved
  BYTE BETransferCount[2];		// Transfer length (big endian)
  BYTE Control;
} CDB_10;

typedef struct
{
  BYTE Opcode;				// operation code
  BYTE LUNRelAddr;			// LUN/Relative address flag
  BYTE BEBlock[4];			// block number (big endian)
  BYTE Reserved;			// reserved
  BYTE BETransferCount[4];		// Transfer length (big endian)
  BYTE Control;
} CDB_12;

typedef union
{
  CDB_6 CDB6;				// six-byte CDB
  CDB_10 CDB10;				// ten-byte CDB
  CDB_12 CDB12;				// twelve-byte CDB
} CDB;

//  Return structure from Inquiry (12h) command.  We don't handle the page-
//  specific items.

typedef struct
{
  BYTE DeviceType;			// device type and qualifier
  BYTE RMB_Reserved;			// reserved = 0, RMB = bit 7
  BYTE Version;				// ANSI-approved version (SCSI-1/2)
  BYTE Response;			// Response flags and format
  BYTE Additional;			// Additional length
  BYTE Reserved_2;			// reserved = 0
  BYTE WideAddress;			// addressing flags (16/32)
  BYTE Options;				// supported options
  CHAR VendorID[8];			// Vendor identification
  CHAR ProductID[16];			// Product identification
  CHAR Revision[4];			// Revision level
  CHAR SerialNo[8];			// Serial number
  BYTE Reserved_3[12];			// Reserved for vendor, zero
  BYTE Reserved_4[40];			// Reserved, zero
  CHAR Copyright[48];			// Copyright notice
  CHAR DistributionSerial[4];		// Distr. Diskette S/N
} SCSI_INQUIRY_RESULTS, *PSCSI_INQUIRY_RESULTS, FAR *LPSCSI_INQUIRY_RESULTS;

#define SCSI_INQ_RMB 128		// RMB bit

//  Return structure from Read Capacity (25h) command.

typedef struct
{
  BYTE BEBlock[4];			// block address (big endian)
  BYTE BELength[4];			// block size (big endian)
} SCSI_CAPACITY_RESULTS, *PSCSI_CAPACITY_RESULTS, FAR *LPSCSI_CAPACITY_RESULTS;

//  Reaturn structure from Read Position (34h) command.

typedef struct
{
  BYTE BEPosition1[3];			// position for SCSI-I
  BYTE Reserved1;
  BYTE BEPosition2[4];			// position for SCSI-II
  BYTE Reserved2[12];			// reserved
} SCSI_READPOS_RESULTS, *PSCSI_READPOS_RESULTS, FAR *LPSCSI_READPOS_RESULTS;

//  Return structure from Read Block Limits (05h) command.

typedef struct
{
  BYTE Reserved;
  BYTE BELargest[3];			// largest block (big endian)
  BYTE BESmallest[2];			// smallest block (big endian)
} SCSI_BLOCKLIM_RESULTS, *PSCSI_BLOCKLIM_RESULTS, FAR *LPSCSI_BLOCKLIM_RESULTS;

//  Header block for Mode Select command.

typedef struct
{
  BYTE	ModeDataLength;			// Must be 0 for mode select
  BYTE	MediumType;			// medium type
  BYTE	Reserved2:4;			// reserved
  BYTE	BufferMode:4;			// buffering mode (upper 4 bits)
  BYTE	BlockDescLen;			// block descriptor length (8)
} SCSI_MODE_SELECT_HEADER, *PSCSI_MODE_SELECT_HEADER,
  FAR *LPSCSI_MODE_SELECT_HEADER;

//  ModeSelect flag in BEBlockLUN[0]

#define SCSI_MODESEL_PF 0x10		// SCSI-2 Mode select format

//  ModeSense flag in BEBlockLUN[0]

#define SCSI_MODESENSE_NOBD 0x08	// No MS block descriptor

//  ModeSense Flags in BEBlockLUN[1].

#define SCSI_GET_DEFAULT_VALUES 0x80	// MS returns default values
#define SCSI_GET_CHANGEABLE_VALUES 0x40 // MS returns changeable values
#define SCSI_GET_CURRENT_VALUES 0x00	// MS returns current values

//  Page overhead in bytes for Mode Sense/Select Pages.	 That is, this
//  quantity must be subtracted from the actual page size before it's
//  stored in the PageLength field of a MS page.

#define SCSI_MS_PAGE_HEADER_SIZE 2	// size in bytes of page header

//  Parameter block for Mode Block Descriptor.

typedef struct
{
  BYTE	Density;			// Density code
  BYTE	BEBlockCount[3];		// logical blocks/physical
  BYTE	BEBlockSize[4];			// block size
} SCSI_MODE_BLOCK_DESC, *PSCSI_MODE_BLOCK_DESC,
  FAR *LPSCSI_MODE_BLOCK_DESC;

//  Parameter blocks for Mode Sense/Select command.

#define SCSI_DRIVE_GEOMETRY_PAGE_TYPE 0x04  // page type

typedef struct tagSCSIDriveGeoPage
{
  BYTE PageCode;			// page code = 4
  BYTE PageLength;			// page length
  BYTE BECylinders[3];			// number of cylinders
  BYTE Heads;				// number of heads
  BYTE BEPrecomp[3];			// write precomp cylinder
  BYTE BERWC[3];			// reduced current cylinder
  BYTE BEStepRate[2];			// step rate
  BYTE BELanding[3];			// landing zone
  BYTE RPL;				// remote position locking
  BYTE Skew;				// rotational skew
  BYTE Reserved_1;			// rezerved-zero
  BYTE BERotation[2];			// medium rotation rate
  BYTE Reserved_2[4];			// reserved-zero
} SCSI_DRIVE_GEO_PAGE, *PSCSI_DRIVE_GEO_PAGE,
  FAR *LPSCSI_DRIVE_GEO_PAGE;

#define SCSI_DEVICE_CONF_PAGE_TYPE 0x10 // page type

typedef struct tagSCSIDeviceConfPage
{
  BYTE	PageCode;			// page code = 10
  BYTE	PageLength;			// page length (always 14 bytes)
  BYTE	ActiveFormat:5;			// active format
  BYTE	CAF:1;				// Change active format
  BYTE	CAP:1;				// change active partition
  BYTE	Reserved1:1;			// reserved; zero
  BYTE	ActivePartition;		// active partition number
  BYTE	WriteFullRatio;			// write buffer full ratio
  BYTE	ReadEmptyRatio;			// read buffer empty ratio
  BYTE	BEWriteDelay[2];		// write-delay (BE)
  BYTE	REW:1;				// report-early warning
  BYTE	RBO:1;				// recover-buffer order
  BYTE	SOCF:2;				// stop on consecutive filemarks
  BYTE	AVC:1;				// automatic velocity control
  BYTE	RSmk:1;				// report setmarks
  BYTE	BIS:1;				// block ID supported
  BYTE	DBR:1;				// data buffer recovery
  BYTE	GapSize;			// write-gap size
  BYTE	Reserved2:3;			// reserved
  BYTE	SEW:1;				// sense early warning
  BYTE	EEG:1;				// enable EOD generation
  BYTE	EOD:3;				// EOD defined
  BYTE	BEEarlyWarnSize[2];		// early warning buffer size
  BYTE	DataCompression;		// select data compression
  BYTE	Reserved3;			// reserved
} SCSI_DEVICE_CONF_PAGE, *PSCSI_DEVICE_CONF_PAGE,
  FAR *LPSCSI_DEVICE_CONF_PAGE;

#define SCSI_MEDIUM_PARTITIONS_PAGE 0x11  // medium partitions page

typedef struct tagSCSIMedPartsPage
{
  BYTE	PageCode;			// page code = 10
  BYTE	PageLength;			// page length (always 6/8 bytes)
  BYTE	MaxAddParts;			// maximum additional partitions (sense)
  BYTE	AddParts;			// additional partitions defined
  BYTE	Reserved:3;			// reserved, 0
  BYTE	PSUM:2;				// partition size units (2=MB)
  BYTE	IDP:1;				// initiator defined partitions
  BYTE	SDP:1;				// select data partitions
  BYTE	FDP:1;				// fixed data partitions
  BYTE	FormatRecognition;		// must be 03h
  BYTE	Reserved2;			// zero
  BYTE	Reserved3;			// zero
  BYTE	BEPartSize[2];			// partition size
} SCSI_MED_PARTS_PAGE, *PSCSI_MED_PARTS_PAGE,
  FAR *LPSCSI_MED_PARTS_PAGE;

typedef struct tagSCSIBlockDescriptor
{
  BYTE	Density;			// density code
  BYTE	BEBlockCount[3];		// block count
  BYTE	Reserved2;			// reserved and zero
  BYTE	BEBlockLength[3];		// block length
} SCSI_BLOCK_DESCRIPTOR, *PSCSI_BLOCK_DESCRIPTOR,
  FAR *LPSCSI_BLOCK_DESCRIPTOR;

typedef struct tagSCSIModeParamHeader
{
  BYTE	SenseLength;			// sense data length to follow
  BYTE	MediumType;			// medium type
  BYTE	Speed:4;			// speed, usually 0
  BYTE	BufferMode:3;			// buffering mode?
  BYTE	WPS:1;				// write-protected
  BYTE	DescLength;			// block descriptor length following
} SCSI_MODE_PARAM_HEADER;

typedef struct tagSCSIModeParamList
{
  SCSI_MODE_PARAM_HEADER Header;	// mode parameter header
  SCSI_BLOCK_DESCRIPTOR	 BlockDesc;	// block descriptor
  SCSI_DEVICE_CONF_PAGE ConfigPage;	// configuration page
} SCSI_MODE_PARAM_LIST;

typedef struct tagSCSIReadCapResults
{
  BYTE BEBlockCount[4];			// count of blocks
  BYTE BEBlockLength[4];		// size of a block
} SCSI_READ_CAPACITY_RESULTS;

typedef struct tagSCSIRequestSenseResults
{
  BYTE ErrorCode:7;			// usually 70 or 71
  BYTE ValidFlag:1;			// set if valid position data
  BYTE SegmentNumber;			// usually 0
  union
  {					// define an alias
    struct
    {
      BYTE SenseKey:4;			    // sense key code
      BYTE Reserved1:1;			    // unused
      BYTE ILI:1;			    // illegal length indicator
      BYTE EOM:1;			    // end-of-mendium
      BYTE FileMark:1;			    // file mark detected
    };
    BYTE SenseByte;			// all 8 bits above
  };
  BYTE BEInformationByte[4];		// information bytes
  BYTE AdditionalSenseLength;		// length of additional sense info
  BYTE BECommandSpecificData[4];	// command specific data
  BYTE AdditionalSenseCode;		// additional sense code
  BYTE AdditionalSenseQualifier;	// additional sense code qualifier
  BYTE FRUCode;				// failing unit code
  BYTE BitPointer:3;			// specific bit pointer
  BYTE BitPointerValid:1;		// nonzero if pointer valid
  BYTE Reserved2:1;			// reserved
  BYTE CommandData:1;			// command/data flag
  BYTE SKSV:1;				// sense-key specific valid flag
  BYTE DriveErrorCode[2];		// drive error code
} SCSI_REQUEST_SENSE_RESULTS, *PSCSI_REQUEST_SENSE_RESULTS,
  FAR *LPSCSI_REQUEST_SENSE_RESULTS;

//  SCSI commands.

typedef enum
{
  SCSI_CMD_TEST_UNIT_READY	= 0x00,
  SCSI_CMD_REZERO_UNIT		= 0x01,
  SCSI_CMD_REWIND		= 0x01,
  SCSI_CMD_REQUEST_SENSE	= 0x03,
  SCSI_CMD_FORMAT_UNIT		= 0x04,
  SCSI_CMD_READ_BLOCK_LIMITS	= 0x05,
  SCSI_CMD_REASSIGN_BLOCKS	= 0x07,
  SCSI_CMD_READ_6		= 0x08,
  SCSI_CMD_RECEIVE		= 0x08,
  SCSI_CMD_WRITE_6		= 0x0a,
  SCSI_CMD_SEND			= 0x0a,
  SCSI_CMD_SEEK_6		= 0x0b,
  SCSI_CMD_SEEK_BLOCK		= 0x0c,
  SCSI_CMD_READ_REVERSE		= 0x0f,
  SCSI_CMD_WRITE_FILEMARKS	= 0x10,
  SCSI_CMD_SPACE		= 0x11,
  SCSI_CMD_INQUIRY		= 0x12,
  SCSI_CMD_RECOVER_BUFFERED_DATA = 0x14,
  SCSI_CMD_MODE_SELECT		= 0x15,
  SCSI_CMD_RESERVE		= 0x16,
  SCSI_CMD_RELEASE		= 0x17,
  SCSI_CMD_COPY			= 0x18,
  SCSI_CMD_ERASE		= 0x19,
  SCSI_CMD_MODE_SENSE		= 0x1a,
  SCSI_CMD_START_STOP		= 0x1b,
  SCSI_CMD_UNLOAD		= 0x1b,
  SCSI_CMD_RECEIVE_DIAGNOSTIC	= 0x1c,
  SCSI_CMD_SEND_DIAGNOSTIC	= 0x1d,
  SCSI_CMD_ALLOW_MEDIUM_REMOVAL = 0x1e
} SCSI_GROUP_0;

typedef enum
{
  SCSI_CMD_READ_CAPACITY	= 0x25,
  SCSI_CMD_READ_10		= 0x28,
  SCSI_CMD_WRITE_10		= 0x2a,
  SCSI_CMD_SEEK_10		= 0x2b,
  SCSI_CMD_WRITE_VERIFY		= 0x2e,
  SCSI_CMD_VERIFY		= 0x2f,
  SCSI_CMD_SEARCH_HIGH		= 0x30,
  SCSI_CMD_SEARCH_EQUAL		= 0x31,
  SCSI_CMD_SEARCH_LOW		= 0x32,
  SCSI_CMD_SET_LIMITS		= 0x33,
  SCSI_CMD_PRE_FETCH		= 0x34,
  SCSI_CMD_READ_POSITION	= 0x34,
  SCSI_CMD_SYNCRONIZE_CACHE	= 0x35,
  SCSI_CMD_LOCK_UNLOCK_CACHE	= 0x36,
  SCSI_CMD_READ_DEFECT_DATA	= 0x37,
  SCSI_CMD_COMPARE		= 0x39,
  SCSI_CMD_COPY_VERIFY		= 0x3a,
  SCSI_CMD_WRITE_BUFFER		= 0x3b,
  SCSI_CMD_READ_BUFFER		= 0x3c,
  SCSI_CMD_READ_LONG		= 0x3e,
  SCSI_CMD_CHANGE_DEFINITION	= 0x40,
  SCSI_CMD_LOG_SELECT		= 0x4c,
  SCSI_CMD_LOG_SENSE		= 0x4d,
  SCSI_CMD_MODE_SELECT_10	= 0x55,
  SCSI_CMD_MODE_SENSE_10	= 0x5a
} SCSI_GROUP_1;

//  SCSI Device types.

typedef enum
{
  SCSI_DEV_DISK = 0,			// RW Disk drive
  SCSI_DEV_TAPE,			// Tape drive
  SCSI_DEV_PRINTER,			// Printer
  SCSI_DEV_PROCESSOR,			// Processor
  SCSI_DEV_WORM,			// WORM drive
  SCSI_DEV_CDROM,			// CDROM drive
  SCSI_DEV_SCANNER,			// Scanner
  SCSI_DEV_OPTICAL,			// Optical drive
  SCSI_DEV_CHANGER,			// Autochanger
  SCSI_DEV_COMM				// Communications
} SCSI_DEVICES;

//    Sense keys.

typedef enum
{
  SCSI_SENSE_NO_SENSE	      = 0x00,
  SCSI_SENSE_RECOVERED_ERROR  = 0x01,
  SCSI_SENSE_NOT_READY	      = 0x02,
  SCSI_SENSE_MEDIUM_ERROR     = 0x03,
  SCSI_SENSE_HARDWARE_ERROR   = 0x04,
  SCSI_SENSE_ILLEGAL_REQUEST  = 0x05,
  SCSI_SENSE_UNIT_ATTENTION   = 0x06,
  SCSI_SENSE_DATA_PROTECT     = 0x07,
  SCSI_SENSE_BLANK_CHECK      = 0x08,
  SCSI_SENSE_FIRMWARE_ERROR   = 0x09,
  SCSI_SENSE_COPY_ABORTED     = 0x0a,
  SCSI_SENSE_ABORTED_COMMAND  = 0x0b,
  SCSI_SENSE_EQUAL	      = 0x0c,
  SCSI_SENSE_VOLUME_OVERFLOW  = 0x0d,
  SCSI_SENSE_MISCOMPARE	      = 0x0e,
  SCSI_SENSE_MASK	      = 0x0f,	// mask for basic status
  SCSI_SENSE_FILEMARK	      = 0x80,	// flag
  SCSI_SENSE_MEDIA_END	      = 0x40,	// flag
  SCSI_SENSE_BAD_LENGTH	      = 0x20	// flag
} SCSI_SENSE_KEY;

#define SCSI_INCLUDED
#endif
