//  Definitions for Tape drives
//  ---------------------------

//  Basic definitions.

#pragma pack(1)

#include "commdefs.h"

//  Basic constants.

#define TAPE_PHYS_BLOCK_SIZE 512U	 // phsyical block size
#define TAPE_BLOCK_SIZE 64000U		 // block size
#define TAPE_BLOCKING_FACTOR (TAPE_BLOCK_SIZE/TAPE_PHYS_BLOCK_SIZE)
#define TAPE_FLAW_BLOCKS 0		 // blocks allocated for flaw map, default
#define TAPE_MAGIC_NUMBER 0x53424454L	 // Tape magic number "SBDT"
#define TAPE_BUFFERS 1			 // how many buffers to use?
#define TAPE_HEADER_VERSION 100		 // header version * 100
#define TAPE_HEADER_CASEID_LENGTH 40	 // size of case number

//  Physical constants.

#define TAPE_RW_TIMEOUT 10		// timeout (seconds) for tape read/write
#define TAPE_POS_TIMEOUT 120		// timeout for positioning commands
#define TAPE_MODE_SENSE_RETRIES 4	// number of mode-sense retries
#define TAPE_REQUEST_SENSE_RETRIES 4	// number of request-sense retries

//  Tape volume table of contents structure.

#define TAPE_VTOC_COUNT 27		// how many VTOC entries

//  Note that the following has a structure identical to that of
//  QS_DRIVE_RETURN.  This saves us a bunch of effort in our code.

typedef struct tagTapeVTOC
{
  BYTE Letter;			// always say what we're talking about
  BYTE AccessFlags;		// nonzero if this drive was direct mode
  SHORT SectorSize;		// size of sector, or -1 if drive error
  LONG TotalSectors;		// megabytes
  SHORT SectorsPerTrack;	// sectors per track
  SHORT TracksPerCylinder;	// tracks per cylinder
  SHORT Cylinders;		// cylinders per drive
} TAPE_VTOC, *PTAPE_VTOC, FAR *LPTAPE_VTOC;

typedef union tagTapeHeader
{
  struct
  {
    DWORD Signature;			  // see TAPE_MAGIC_NUMBER, above
    WORD  HeaderSize;			  // header size in bytes
    WORD  HeaderCRC;			  // CRC-16 of header
    WORD  Volume;			  // what volume ID?
    WORD  Sequence;			  // sequence number
    WORD  Flags;			  // flags; unused are zero
    WORD  BlockSize;			  // tape block size in Kbytes
    WORD  MapBlocks;			  // how many blocks in map?
    WORD  Version;			  // version of this header
    DWORD DateTime;			  // date/time (time_t format)
    CHAR  HardwareID[16];		  // Hardware model number
    CHAR  VendorID[8];
    CHAR  Product[40];			  // product ID (SafeBack x.xx
    CHAR  CaseNumber[TAPE_HEADER_CASEID_LENGTH];  // case number?
    CHAR  Author[64];			  // author ID
    TAPE_VTOC VTOC[ TAPE_VTOC_COUNT];	  // VTOC's
  };
  CHAR	Reserved[2*TAPE_PHYS_BLOCK_SIZE];    // for clearing out
} TAPE_VOLUME_HEADER, *PTAPE_VOLUME_HEADER, FAR *LPTAPE_VOLUME_HEADER;

//  The tape hook structure for TapeInit() and other functions.

typedef struct tagTapeInfo
{
  LPTAPE_VOLUME_HEADER Header;	    // pointer to block header
  LPVOID Buffer;		    // tape block buffer
  LPASPI_SRB SRB;		    // ASPI SRB pointer
  DWORD BlockNumber;		    // current written block
  SHORT Volume;			    // current volume number
  SHORT Function;		    // what SCSI function?
  SHORT OpenMode;		    // open mode
  BYTE ErrorCode;		    // error code, if any
  BYTE SenseKey;		    // sense key, if any
  SHORT	 (FAR *PreMount)(struct tagTapeInfo  FAR *Info);
  SHORT	 (FAR *PostMount)(struct tagTapeInfo  FAR *Info);
  VOID (FAR *EndTape)(struct tagTapeInfo FAR *Info);
  SHORT	 (FAR *TapeError)(struct tagTapeInfo FAR *Info);
} TAPE_INFO, *PTAPE_INFO, FAR *LPTAPE_INFO;

//  ErrorCode values for the above structure.

typedef enum
{
  TAPE_ERROR_NONE	= 0,		// no error
  TAPE_ERROR_HEADER	= -1,		// header problem
  TAPE_ERROR_EOT	= -2,		// EOT error
  TAPE_ERROR_PROTECTED	= -3,		// tape is write protected
  TAPE_ERROR_NOTREADY	= -4,		// drive not ready
  TAPE_ERROR_ABORT	= -5,		// user aborted operation
  TAPE_ERROR_MEMORY	= -6,		// no memory on init
  TAPE_ERROR_ILLEGAL	= -7,		// illegal operation
  TAPE_ERROR_VOLUME	= -8,		// wrong volume mounted
  TAPE_ERROR_DATA	= -9,		// tape data error
  TAPE_ERROR_BLANK	= -10,		// unexpected blank tape
  TAPE_ERROR_EOF	= -11		// EOF in buffer (informational)
} TAPE_ERRORS;

//  Prototypes.

#define TAPE_READ_MODE 0		// for mount; mount for read
#define TAPE_WRITE_MODE 1		// for mount; mount for write
#define TAPE_QUERY_MODE 2		// for mount; query mount mode
#define TAPE_QUERY_VOLUME 3		// for mount; query volume number
#define TAPE_UNLOAD_MODE 4		// for mount; unload the tape

SHORT FAR IsTape( LPCHAR ProductID, LPCHAR VendorID,
		 LPDWORD maxblock);
SHORT FAR TapeInit( SHORT (FAR *PreMount)(LPTAPE_INFO),
		   SHORT (FAR *PostMount)(LPTAPE_INFO),
		   VOID (FAR *EndTape)(LPTAPE_INFO),
		   SHORT (FAR *TapeError)(LPTAPE_INFO));
SHORT FAR TapeMount( SHORT Volume,SHORT mode);
LPTAPE_INFO TapeUpdateHeader( SHORT mode);
SHORT FAR TapeWrite( LPVOID buf, LONG count);
SHORT FAR TapeSkip( SHORT blocks);
LONG FAR TapeRead( LPVOID buf, LONG count);
SHORT FAR TapeTestReady( SHORT mode);
