//  Tape I/O for DAT copy program - tapeio.c.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "tapeio.h"

#define MAX_TAPE_DRIVES 7		// how many tape drives

typedef struct _TAPE_DRIVE
{

  BYTE Adapter;				// Adapter ID of tape
  BYTE TargetID;			// Target ID of tape
  BYTE SCSI2;				// nonzero if at least SCSI2
  WORD BlockSize;			// block size
  DWORD Position;			// block position
} TAPE_DRIVE, *PTAPE_DRIVE;

TAPE_DRIVE
  Tape[ MAX_TAPE_DRIVES];
int
  TapeCount;

static union
{
  ASPI_SRB	Request;	   // basic SRB header
  ASPI_INQUIRE	Inquire;	   // inquire
  ASPI_GETTYPE	GetType;	   // get type
  ASPI_EXECUTE	Execute;	   // execute command
  ASPI_ABORT	Abort;		   // abort command
  ASPI_RESET	Reset;		   // reset device
  ASPI_SETPARAM SetParam;	   // set device parameters
  ASPI_GETINFO	GetInfo;	   // get device information
} SRB;

#define CMD6 SRB.Execute.CommandBlock.CDB6  // alias for short CDB
#define CMD10 SRB.Execute.CommandBlock.CDB10 // alias for long CDB
#define CLEAR(x) _fmemset( addr(x), 0, sizeof(x))

//  Local Routines.

__inline static void ClearSCSIStatus( void);
__inline static void ClearSRB( void);
__inline static void ClearSCSIBlock( void);

//  GetTapes - Is there a tape device?
//  --------------------------------
//
//    Polls ASPI interface.  Returns TRUE if so, FALSE if not.
//
//    Note that NULL may be passed in for both product and vendor ID pointers.
//    If the block limits are requested, we'll send them too.
//
//    Returns 1 if scan successful, or 0 if no ASPI present


int GetTapes( void)
{


  int
    AdapterCount,			// total adapters
    stat,				// general status
    i,j;				// index variables

  SCSI_INQUIRY_RESULTS
    InqResults;				// SCSI inquiry results

  SCSI_REQUEST_SENSE_RESULTS
    SenseResults;			// SCSI Request sense results

  CHAR
   VendorID[8+1],			// Vendor identification
   ProductID[16+1],			// Product identification
   Revision[4+1],			// Revision level
   SerialNo[8+1],			// Serial number
   Copyright[48+1],			// Copyright notice
   DistributionSerial[4+1];		// Distr. Diskette S/N

  TapeCount = 0;			// say no tapes yet

  if ( !AccessASPI())
    return -1;			       // no soap right away

//  Okay we have at least one adapter--find out how many we really do have.

  if ( CallASPI( addr(SRB.Request), ASPI_CMD_INQUIRE, 0, 0))
    return FALSE;
  AdapterCount = SRB.Inquire.Adapters;
  for ( i = 0; i < AdapterCount; i++)
  {
    if ( !CallASPI( addr(SRB.Request), ASPI_CMD_INQUIRE, (BYTE) i, 0))
    {
      for ( j = 0; j < 8; j++)
      { // for each ID
	SRB.GetType.TargetID = (BYTE) j;
	SRB.GetType.LUN = 0;
	if ( !CallASPI( addr(SRB.GetType), ASPI_CMD_GETTYPE, (BYTE) i, 0))
	{ // a good get
	  if ( SRB.GetType.DeviceType == SCSI_DEV_TAPE)
	  { // we found it!
	    Tape[ TapeCount].Adapter = (BYTE) i;
	    Tape[ TapeCount].TargetID = (BYTE) j;
	    Tape[ TapeCount].BlockSize = 1; // assume unblocked

//  Do an enquire here to get product information.  Do a test ready to
//  clear the unit attention status.
 
	    TapeTestReady(0,TapeCount);
	    TapeRequestSense( addr(SenseResults), TapeCount);
	    ClearSRB();
	    _fmemset( addr(InqResults), ' ', sizeof(InqResults));
	    CMD6.TransferCount = sizeof( InqResults);
	    if ( !(stat = ExecASPI( addr(SRB.Execute), SCSI_CMD_INQUIRY,
		 (BYTE) i, 0, (BYTE) j, addr(InqResults),
		 sizeof(SCSI_INQUIRY_RESULTS)) ) )
	    { // got some product data
	      memcpy( VendorID, InqResults.VendorID,
			sizeof(InqResults.VendorID));
	      VendorID[ sizeof(VendorID)-1] = 0; // null
	      memcpy( ProductID, InqResults.ProductID,
			sizeof(InqResults.ProductID));
	      ProductID[ sizeof(ProductID)-1] = 0;
	      strncpy( SerialNo, InqResults.SerialNo,
			sizeof(InqResults.SerialNo));
	      SerialNo[ sizeof(SerialNo) -1] = 0;
	      Tape[ TapeCount].SCSI2 = FALSE;
	      if ( (InqResults.Version & 7) >= 2)
		Tape[ TapeCount].SCSI2 = TRUE;	     // yes--it's at least SCSI2
	      TapeCount++;
	      if ( TapeCount == 1)
		printf( "\nThe following SCSI tape devices were found:\n\n"
		 "No.  Adapt.     ID     Vendor        Product            S/N\n"
		 "---  ------     --     ------        -------            ---\n");
	      printf( "%2d      %d       %d     %8s     %16s   %s\n",
		TapeCount, i, j, VendorID, ProductID,SerialNo);
	      if ( TapeCount >= MAX_TAPE_DRIVES)
		break;
	    } // if Inquire worked
	  } // if we found a tape unit
	} // if we succeeded in a Get Type
      } //  for each ID
    } // if the inquire worked
      if ( TapeCount >= MAX_TAPE_DRIVES)
	break;
  } // for each adapter
  return TapeCount;			  // We found something
} // GetTapes

//  ClearSCSIBlock - Clear SCSI Command Block
//  -----------------------------------------
//
//    Clears out 12 bytes of the SCSI command block.
//

__inline static void ClearSCSIBlock( void)
{
  _fmemset( addr(SRB.Execute), 0, sizeof( ASPI_EXECUTE));
  SRB.Request.Status = ASPI_STATUS_NONE;
  return;
} // ClearSCSIBlock

//  ClearSRB - Clear ASPI Command Block
//  -----------------------------------
//
//    Clears out 12 bytes of the SCSI command block.
//

__inline static void ClearSRB( void)
{
  _fmemset( addr(SRB.Execute), 0, sizeof( SRB));
  return;
} // ClearSRB

//  TapeRequestSense - Do a Mode-sense on the drive.
//  ---------------------------------------------
//
//  There are some transient conditions on the drive, such as media-change
//  which require two Request-Sense commands.  This routine takes care of the
//  requirement.
//

int TapeRequestSense( SCSI_REQUEST_SENSE_RESULTS *rsp, int TapeNo)
{

  int
    status,
    retries;

  for (retries = 0; retries < TAPE_REQUEST_SENSE_RETRIES; retries++)
  {
    _fmemset( rsp, 0, sizeof(SCSI_REQUEST_SENSE_RESULTS)); // clear descriptor
    ClearSRB();			    // clear the command block
    CMD6.TransferCount = sizeof( SCSI_REQUEST_SENSE_RESULTS);
    if ( !ExecASPI( addr( SRB.Execute), SCSI_CMD_REQUEST_SENSE,
	      Tape[ TapeNo].Adapter, 0,
	      Tape[ TapeNo].TargetID,
	      rsp, sizeof(SCSI_REQUEST_SENSE_RESULTS)))
    {
      status =	SCSI_SENSE_NO_SENSE;	   // say it's okay
      break;
    }
    else
      status = ASPIGetSenseByte( addr(SRB.Execute));
  } // for getting sense data
  return status;
} // TapeRequestSense

//  TapeTestReady - Test the tape unit for ready.
//  ---------------------------------------------
//
//    Handles retry operations.	 If mode is nonzero, we intend to write
//    on this tape.
//
//    If TAPE_QUERY_MODE is passed, no error hook is done, but -1 is
//    returned if drive not ready.
//

int TapeTestReady(  int mode, int TapeNo)
{

  SCSI_MODE_PARAM_LIST
    ModeSense;				// mode sense parameters

//  First of all, check for ready.

  for(;;)
  {
    for(;;)
    {
      ClearSRB();		      // clear the command block
      if ( !ExecASPI( addr( SRB.Execute), SCSI_CMD_TEST_UNIT_READY,
		      Tape[ TapeNo].Adapter, 0, Tape[ TapeNo].TargetID,
		      NULL, 0))
	break;				// the drive is ready
      else
      {
	ClearSCSIStatus();
	return SCSI_SENSE_NOT_READY;
      }
    } // for
    if ( !mode )
      return SCSI_SENSE_NO_SENSE;	    // it's okay

//  Next, check for write enabled.

    CLEAR( ModeSense);			  // clear descriptor
    ClearSRB();			    // clear the command block
    CMD6.TransferCount = sizeof(SCSI_MODE_PARAM_LIST);
    if ( ExecASPI( addr( SRB.Execute), SCSI_CMD_MODE_SENSE,
		Tape[ TapeNo].Adapter, 0,
		Tape[ TapeNo].TargetID, addr(ModeSense),
		 sizeof(SCSI_MODE_PARAM_LIST)) )
      return SCSI_SENSE_ILLEGAL_REQUEST;
    if ( ModeSense.Header.WPS)
	return SCSI_SENSE_DATA_PROTECT; // return protected
    else
      return SCSI_SENSE_NO_SENSE;	// say it's okay
  } // forever
} // TapeTestReady


//  Rewind - Rewind a tape.
//  -----------------------
//
//    This function never waits for I/O.
//

int TapeRewind( int TapeNo)
{

  SetASPITime( TAPE_POS_TIMEOUT);	// could take this long for rewind
  ClearSRB();
  ExecASPI( addr( SRB.Execute), SCSI_CMD_REZERO_UNIT, Tape[ TapeNo].Adapter,
	    0, Tape[ TapeNo].TargetID, NULL, 0 );
  Tape[ TapeNo].Position = 0;		// back to start
  return TapeClearIO( TapeNo);			  // wait for I/O to finish
} // Rewind

//  Unload - Rewind a tape.
//  -----------------------
//
//    This function never waits for I/O.
//

int TapeUnload( int TapeNo)
{

  SetASPITime( TAPE_POS_TIMEOUT);	// could take this long for rewind
  ClearSRB();
  ExecASPI( addr( SRB.Execute), SCSI_CMD_START_STOP, Tape[ TapeNo].Adapter,
	    0, Tape[ TapeNo].TargetID, NULL, 0 );
  Tape[ TapeNo].Position = 0;		// back to start
  return 0;		       // wait for I/O to finish
} // Rewind


//  TapeReadPosition - Read current tape block.
//  -------------------------------------------
//
//    Returns with a complete status.
//

DWORD TapeReadPosition( int TapeNo)
{

  DWORD
    retval;				// return value
  SCSI_READPOS_RESULTS
    ReadPos;				// return block for read position

  ClearSRB();
  ExecASPI( addr( SRB.Execute), SCSI_CMD_READ_POSITION,
	    Tape[ TapeNo].Adapter, ASPI_EXEC_DIR_READ,
	    Tape[ TapeNo].TargetID, addr(ReadPos), sizeof( ReadPos));

  if ( Tape[ TapeNo].SCSI2)
    retval = ASPIGet4( ReadPos.BEPosition2);
  else
    retval = ASPIGet3( ReadPos.BEPosition1);
  return retval;
} // TapeReadPosition

//  TapeSkipFiles - Skip Filemarks.
//  -------------------------------
//
//    Returns with completion status.
//

int TapeSkipFiles( long Position, int TapeNo)
{

  ClearSRB();
  CMD6.LUN = 1;				// say skip filemarks
  ASPIPut3( CMD6.BETransferLength, Position);
  ExecASPI( addr( SRB.Execute), SCSI_CMD_SPACE,
	    Tape[ TapeNo].Adapter, 0, Tape[ TapeNo].TargetID, NULL, 0);
  return TapeClearIO(TapeNo);			 // wait for I/O to finish
} // TapeSkipFiles

//  TapeSetBlock - Set Tape Block Size.
//  -----------------------------------
//
//    Sets default density and block size as specifed.	If writing, sets
//    partition size.
//

int TapeSetBlock( WORD TapeBlockSize, int Mode, int TapeNo)
{

  SCSI_MODE_PARAM_LIST
    ModeSense;				// mode sense parameters

  struct
  {
    SCSI_MODE_SELECT_HEADER
      SelectHeader;		       // mode select parameters
    SCSI_MODE_BLOCK_DESC
      BlockDesc;		       // block descriptor
  } BlockLimPage;

  struct
  {
    SCSI_MODE_SELECT_HEADER
      SelectHeader;			// mode select parameters
    SCSI_MED_PARTS_PAGE
      MediaParts;			// media partitions page
  } PartitionPage;

//  Set the physical block length, buffering and density.

  if( Tape[ TapeNo].BlockSize == TapeBlockSize)
    return SCSI_SENSE_NO_SENSE;		// ignore if already set
  TapeTestReady( TapeNo, TAPE_QUERY_MODE);
  CLEAR( ModeSense);			// clear descriptor
  ClearSRB();			  // clear the command block
  CMD6.TransferCount = sizeof( SCSI_MODE_PARAM_LIST);
  if ( ExecASPI( &SRB.Execute, SCSI_CMD_MODE_SENSE,
		 Tape[ TapeNo].Adapter, ASPI_EXEC_DIR_UNSPEC,
		 Tape[ TapeNo].TargetID, &ModeSense,
		 sizeof(SCSI_MODE_PARAM_LIST)) )
    return SCSI_SENSE_ILLEGAL_REQUEST;
  else
  { // if sense worked
    CLEAR( BlockLimPage);
    ClearSRB();
    BlockLimPage.SelectHeader.BufferMode = 1;  // set buffering mode
    BlockLimPage.SelectHeader.BlockDescLen = 8;	  // descriptor length
    ASPIPut4( BlockLimPage.BlockDesc.BEBlockSize, TapeBlockSize);
    BlockLimPage.BlockDesc.Density = 0;	 // set density (default)
    CMD6.TransferCount = sizeof(BlockLimPage);
    SetASPITime( TAPE_POS_TIMEOUT);
    if ( ExecASPI( &SRB.Execute, SCSI_CMD_MODE_SELECT,
		  Tape[ TapeNo].Adapter, ASPI_EXEC_DIR_UNSPEC,
		  Tape[ TapeNo].TargetID, &BlockLimPage,
		  sizeof(BlockLimPage))	 )
      return SCSI_SENSE_ILLEGAL_REQUEST;
    Tape[ TapeNo].BlockSize = TapeBlockSize;
 
//  If writing, set the partition size.

    if ( Mode == TAPE_WRITE_MODE)
    {
      CLEAR( PartitionPage );
      ClearSRB();
      CMD6.TransferCount = sizeof( PartitionPage);
      CMD6.BEBlockLUN[0] = SCSI_MODESENSE_NOBD;
      CMD6.BEBlockLUN[1] = SCSI_GET_DEFAULT_VALUES |
			   SCSI_MEDIUM_PARTITIONS_PAGE;
      SetASPITime( TAPE_RW_TIMEOUT);
      if ( !ExecASPI( &SRB.Execute, SCSI_CMD_MODE_SENSE,
		  Tape[ TapeNo].Adapter, ASPI_EXEC_DIR_UNSPEC,
		  Tape[ TapeNo].TargetID, &PartitionPage,
		  sizeof(PartitionPage))  )
      {
	PartitionPage.SelectHeader.BufferMode = 1;  // set buffering mode
	PartitionPage.SelectHeader.BlockDescLen = 0;   // descriptor length
	PartitionPage.SelectHeader.ModeDataLength = 0;	// clear data length
	ClearSRB();
	CMD6.TransferCount = sizeof(PartitionPage);
	CMD6.BEBlockLUN[0] = SCSI_MODESEL_PF; // say SCSI-2 format
	PartitionPage.MediaParts.PageCode = SCSI_MEDIUM_PARTITIONS_PAGE;
	PartitionPage.MediaParts.AddParts = 0;
	PartitionPage.MediaParts.IDP = 1;   // define initator partitions
	PartitionPage.MediaParts.BEPartSize[0] =
	PartitionPage.MediaParts.BEPartSize[1] = 0;
	PartitionPage.MediaParts.PageLength =
	    sizeof( PartitionPage.MediaParts) - SCSI_MS_PAGE_HEADER_SIZE;
	SetASPITime( TAPE_POS_TIMEOUT);
	ExecASPI( &SRB.Execute, SCSI_CMD_MODE_SELECT,
		  Tape[ TapeNo].Adapter, ASPI_EXEC_DIR_UNSPEC,
		  Tape[ TapeNo].TargetID, &PartitionPage,
		  sizeof(PartitionPage));
      } // if mode-sense worked
      ClearSRB();		  // We don't care about the outcome
    } // if writing
  } // if sense worked
  return SCSI_SENSE_NO_SENSE;
} // TapeSetBlock

//  TapeGetBlockSize - Determine Tape Block Size
//  --------------------------------------------
//
//    Returns block size in bytes.  The tape must be READY for this to work.
//    If failure, returns 000.
//
//    This attempts a verify of 1 byte.	 If ILI gets set, then the residue will
//    give us the -true block size -1.
//
//    We then backspace one record to put us back where we need to be.
//

WORD TapeGetBlockSize( int TapeNo)
{

  BYTE
    dummyBuf;

  LPSCSI_REQUEST_SENSE_RESULTS
    sensePtr;

  LONG
    residual;

  WORD
    newRecordLength;

  newRecordLength = 0;
  if ( !TapeTestReady( 0, TapeNo))
  {

    ClearSRB();
    CMD6.TransferCount = 1;
    ExecASPI( &SRB.Execute, SCSI_CMD_READ_6,
		Tape[ TapeNo].Adapter, ASPI_EXEC_DIR_READ,
		Tape[ TapeNo].TargetID, &dummyBuf, 1);

//  Okay, ILI should be set; if not; the block size is 1 byte!

    sensePtr = ASPIGetSenseInfo( &SRB.Execute);
    if ( sensePtr->SenseKey == SCSI_SENSE_NO_SENSE)
    {
      if ( sensePtr->ILI)
      { // invalid length record
	residual = (LONG) ASPIGet4( sensePtr->BEInformationByte);
	newRecordLength = (WORD) (1L - residual);
      }
      else
	newRecordLength = 1;
    }

//  Regardless, backspace 1 record.

    ClearSRB();
    TapeSkipRelative( -1L, TapeNo);
    ClearSCSIBlock();
  } // if ready
  return newRecordLength;
} // TapeGetBlockSize

//  TapeSkipRelative - Skip Blocks Relative
//  ---------------------------------------
//
//  Returns status.
//

int TapeSkipRelative( long Blocks, int TapeNo)
{

  ClearSRB();
  ASPIPut3( CMD6.BETransferLength, (DWORD) Blocks);

  ExecASPI( addr( SRB.Execute), SCSI_CMD_SPACE,
	    Tape[ TapeNo].Adapter, 0, Tape[ TapeNo].TargetID, NULL, 0);
  return TapeClearIO(TapeNo);
} // TapeSkipRelative

//  TapeErase - Erase Tape.
//  -----------------------
//
//    Just takes the default action.
//

int TapeErase( int TapeNo)
{
  ClearSRB();
  SetASPITime( TAPE_POS_TIMEOUT);
  ExecASPI( addr( SRB.Execute), SCSI_CMD_ERASE, Tape[ TapeNo].Adapter,
	    0, Tape[ TapeNo].TargetID, NULL, 0);
  return TapeClearIO(TapeNo);
} // TapeErase


//  TapeRead - Read blocks of data from tape.
//  -----------------------------------------
//
//    Returns the requested data.  If a short read is encountered,
//    returns the amount actually read.	 If an error occurs, returns
//    the sense key.
//
//    Note that the read must be in multiples of the block size.
//

int TapeRead( void _far *Buffer, WORD BufferSize, WORD *BytesRead, int TapeNo)
{


  BYTE
    SenseByte;
  DWORD
    TapePos;
  WORD
    BlocksRead;


  ClearSRB();
  CMD6.TransferCount = (BYTE) (BufferSize / Tape[ TapeNo].BlockSize);
  if ( Tape[ TapeNo].BlockSize != 1)
    CMD6.BEBlockLUN[0] = 1;	       // say fixed-length
  SetASPITime( TAPE_POS_TIMEOUT);	// could take this long for rewind

  ExecASPI( addr( SRB.Execute), SCSI_CMD_READ_6,
		Tape[ TapeNo].Adapter, ASPI_EXEC_DIR_READ,
		Tape[ TapeNo].TargetID, Buffer,
		BufferSize);
  SenseByte = ASPIGetSenseByte( addr(SRB.Execute));

//  If we have a filemark, get the current position and compute the
//  amount of data transfered.

  if ( SenseByte & SCSI_SENSE_FILEMARK)
  {
    TapePos = TapeReadPosition( TapeNo);
    BlocksRead = (WORD) (TapePos - Tape[TapeNo].Position -1);
    Tape[ TapeNo].Position = TapePos;
    *BytesRead = (WORD) BlocksRead * Tape[ TapeNo].BlockSize;
    if ( SenseByte & SCSI_SENSE_FILEMARK)
      SenseByte = SCSI_SENSE_FILEMARK;
  }
  else
  {
    *BytesRead = BufferSize;
    Tape[ TapeNo].Position += (BufferSize / Tape[TapeNo].BlockSize);
  }
  return SenseByte;
} // TapeRead

//  TapeReadVariable - Read a variable length block.
//  ------------------------------------------------
//
//  Returns the actual number of bytes read.
//
//    Returns block size in bytes.  The tape must be READY for this to work.
//    If failure, returns 000.
//
//    This attempts a verify of 1 byte.	 If ILI gets set, then the residue will
//    give us the -true block size -1.
//
//    We then backspace one record to put us back where we need to be.
//

int TapeReadVariable( void _far *Buffer, WORD BufferSize,
		      WORD *BytesRead, int TapeNo)
{

  LPSCSI_REQUEST_SENSE_RESULTS
    sensePtr;

  WORD
    residual;

  DWORD
    tapePos;

  WORD
    newRecordLength;
 
  BYTE
    senseKey;

  newRecordLength = 0;
  senseKey = SCSI_SENSE_NOT_READY;

  if ( !TapeTestReady( 0, TapeNo))
  {
    ClearSRB();
    ASPIPut3( CMD6.BETransferLength, BufferSize);
    ExecASPI( &SRB.Execute, SCSI_CMD_READ_6,
		Tape[ TapeNo].Adapter, ASPI_EXEC_DIR_READ,
		Tape[ TapeNo].TargetID, Buffer, BufferSize);

//  Okay, ILI should be set; if not; the block size is 1 byte!

    sensePtr = ASPIGetSenseInfo( &SRB.Execute);
    senseKey = sensePtr->SenseKey;

//  Okay, we look for a "NO SENSE" key report, and break it out from there.

    if ( senseKey == SCSI_SENSE_NO_SENSE)
    {
      if ( sensePtr->ILI)
      { // invalid length record
	residual = (WORD) ASPIGet4( sensePtr->BEInformationByte);
	newRecordLength = (WORD) (BufferSize - residual);
	senseKey = SCSI_SENSE_NO_SENSE;
      }
      if ( sensePtr->FileMark)
      {
	tapePos = TapeReadPosition( TapeNo);
	Tape[ TapeNo].Position = tapePos;
	senseKey = SCSI_SENSE_FILEMARK;
      }
      if ( newRecordLength < 2)
	senseKey = SCSI_SENSE_FILEMARK;	  // 0 bytes must mean filemark
    } // if no error
    else
    {
      if ( senseKey == SCSI_SENSE_RECOVERED_ERROR)
      {
	printf( "\nRecovered error\n");
	senseKey = SCSI_SENSE_NO_SENSE;
      } // if a recovered error
    } // other sense key values
  } // if ready
  *BytesRead = newRecordLength;		// return block length
  return senseKey;
} // TapeReadVariable

//  TapeWrite - Write a block of data to tape.
//  ------------------------------------------
//
//    Note that the buffer is written as-is, and must be a multiple
//    of the block size.  If NULL is passed as the buffer pointer,
//    BufferSize filemarks are written.
//
//    The sense key is passed if there is an error.
//

int TapeWrite( void _far *Buffer, WORD BufferSize, int TapeNo)
{

  int
    status;

  LPSCSI_REQUEST_SENSE_RESULTS
    sensePtr;

  BYTE
    blocksToWrite,
    residual;

  if ( !Buffer)
  { //	Write a filemark, if requested.
    ClearSRB();
    CMD6.TransferCount = (BYTE) BufferSize;   // filemark count
    ExecASPI( addr( SRB.Execute), SCSI_CMD_WRITE_FILEMARKS,
	      Tape[ TapeNo].Adapter, 0, Tape[ TapeNo].TargetID, NULL, 0 );

    status = TapeClearIO(TapeNo);
  }
  else
  { // Write Data
    ClearSRB();
    CMD6.TransferCount =
      blocksToWrite = (BYTE) (BufferSize / Tape[TapeNo].BlockSize);
    if ( Tape[ TapeNo].BlockSize != 1)
      CMD6.BEBlockLUN[0] = 1;		 // say fixed-length
    ExecASPI( addr( SRB.Execute), SCSI_CMD_WRITE_6,
	      Tape[ TapeNo].Adapter, ASPI_EXEC_DIR_WRITE,
	      Tape[ TapeNo].TargetID, Buffer, BufferSize);
    status = TapeClearIO(TapeNo);
    if ( status == SCSI_SENSE_NO_SENSE)
    { // need to check for early EOM and possible unwritten data
      sensePtr = ASPIGetSenseInfo( addr(SRB.Execute));
      if ( sensePtr->EOM)
      { // early warning
      residual = (BYTE) ASPIGet4( sensePtr->BEInformationByte);
	if ( residual)
	{ // finish out the write anyway in the EW area.
	  ClearSRB();
	  CMD6.TransferCount = residual;
	  CMD6.BEBlockLUN[0] = 1;	     // say fixed-length
	  ExecASPI( addr( SRB.Execute), SCSI_CMD_WRITE_6,
		    Tape[ TapeNo].Adapter, ASPI_EXEC_DIR_WRITE,
		    Tape[ TapeNo].TargetID,
		    (LPBYTE) Buffer +
		       ((blocksToWrite - residual) *
		       Tape[ TapeNo].BlockSize),
		       residual * Tape[TapeNo].BlockSize);
	  status = TapeClearIO(TapeNo);
	} // if residual
      } // if early warning
    } // if good status
  } // if writing data
  return status;
} // TapeWrite


//  TapeClearIO - Clear any I/O activity on a tape drive.
//  -----------------------------------------------------
//
//    Waits until all I/O is halted.  Returns error status from the last
//    I/O operation, if any.
//

int TapeClearIO( int TapeNo)
{

  int status;

//  Finish any outstanding request.

  if ( SRB.Request.Status == ASPI_STATUS_BUSY)
    while ( PollASPI( addr( SRB.Execute)) == 1);  // wait for completion

//  Handle error status.

  status = SCSI_SENSE_NO_SENSE;
  if ( SRB.Request.Status != ASPI_STATUS_NONE)
  { // whoops -- we have to handle possible error
    status = ASPIGetSenseByte( addr(SRB.Execute));
    ResetASPI( addr(SRB.Reset),
	       Tape[ TapeNo].Adapter, Tape[ TapeNo].TargetID);
  } // if

  SRB.Request.Status = ASPI_STATUS_NONE;
  return status;
} // TapeClearIO


//  ClearSCSIStatus - Clear SCSI Command Status
//  -------------------------------------------
//
//    Clears out the status bytes.
//

__inline static void ClearSCSIStatus( void)
{
  SRB.Request.Status = ASPI_STATUS_NONE;
  return;
} // ClearSCSIStatus

